@@ -6,7 +6,6 @@
 #include <ctime>
 #include <utility>
 #include "Utils.h"
-#include "rapidjson/document.h"
 #include "rapidjson/writer.h"
 #include "rapidjson/stringbuffer.h"
 #include <kodi/Filesystem.h>
@@ -21,41 +20,15 @@ using namespace rapidjson;
 constexpr char app_token_file[] = "special://temp/zattoo_app_token";
 const char data_file[] = "special://profile/addon_data/pvr.zattoo/data.json";
 
-std::string ZatData::GetManifestType()
-{
-  switch (m_settings->GetStreamType())
-  {
-    case HLS:
-      return "hls";
-    default:
-      return "mpd";
-  }
-}
-
-std::string ZatData::GetMimeType()
-{
-  switch (m_settings->GetStreamType())
-  {
-    case HLS:
-      return "application/x-mpegURL";
-    default:
-      return "application/xml+dash";
-  }
-}
-
 void ZatData::SetStreamProperties(
     std::vector<kodi::addon::PVRStreamProperty>& properties,
     const std::string& url)
 {
   properties.emplace_back(PVR_STREAM_PROPERTY_STREAMURL, url);
   properties.emplace_back(PVR_STREAM_PROPERTY_INPUTSTREAM, "inputstream.adaptive");
-  properties.emplace_back("inputstream.adaptive.manifest_type", GetManifestType());
-  properties.emplace_back(PVR_STREAM_PROPERTY_MIMETYPE, GetMimeType());
-
-  if (m_settings->GetStreamType() == DASH || m_settings->GetStreamType() == DASH_WIDEVINE)
-  {
-    properties.emplace_back("inputstream.adaptive.manifest_update_parameter", "full");
-  }
+  properties.emplace_back("inputstream.adaptive.manifest_type", "mpd");
+  properties.emplace_back(PVR_STREAM_PROPERTY_MIMETYPE, "application/xml+dash");
+  properties.emplace_back("inputstream.adaptive.manifest_update_parameter", "full");
 }
 
 bool ZatData::ReadDataJson()
@@ -139,35 +112,39 @@ bool ZatData::LoadChannels()
   for (Value::ConstValueIterator itr1 = channels.Begin();
       itr1 != channels.End(); ++itr1)
   {
+    
     const Value& channelItem = (*itr1);
+    
+    ZatChannel channel;
+    std::string cid = Utils::JsonStringOrEmpty(channelItem, "cid");
+    channel.iUniqueId = Utils::GetChannelId(cid.c_str());
+    channel.cid = cid;
+    channel.iChannelNumber = ++channelNumber;
+    channel.recordingEnabled =
+        channelItem.HasMember("recording") ?
+            channelItem["recording"].GetBool() : false;
+    
     const Value& qualities = channelItem["qualities"];
     for (Value::ConstValueIterator itr2 = qualities.Begin();
         itr2 != qualities.End(); ++itr2)
     {
       const Value& qualityItem = (*itr2);
       std::string avail = Utils::JsonStringOrEmpty(qualityItem, "availability");
-      if (avail == "available")
-      {
-        ZatChannel channel;
+      if (avail != "available") {
+        continue;
+      }
+      bool drmRequired = qualityItem.HasMember("drm_required") ? qualityItem["drm_required"].GetBool() : false;
+      channel.qualityWithDrm.push_back({Utils::JsonStringOrEmpty(qualityItem, "level"), drmRequired});
+      if (channel.name.empty()) {
         channel.name = Utils::JsonStringOrEmpty(qualityItem, "title");
-        std::string cid = Utils::JsonStringOrEmpty(channelItem, "cid");
-        channel.iUniqueId = Utils::GetChannelId(cid.c_str());
-        channel.cid = cid;
-        channel.iChannelNumber = ++channelNumber;
-        channel.strLogoPath = "http://logos.zattic.com";
-        channel.strLogoPath.append(
-            Utils::JsonStringOrEmpty(qualityItem, "logo_white_84"));
-        channel.recordingEnabled =
-            channelItem.HasMember("recording") ?
-                channelItem["recording"].GetBool() : false;
-        PVRZattooChannelGroup &group = m_channelGroups[channelItem["group_index"].GetInt()];
-        group.channels.insert(group.channels.end(), channel);
-        allChannels[cid] = channel;
-        m_channelsByCid[channel.cid] = channel;
-        m_channelsByUid[channel.iUniqueId] = channel;
-        break;
+        channel.strLogoPath = "http://logos.zattic.com" + Utils::JsonStringOrEmpty(qualityItem, "logo_white_84");
       }
     }
+    PVRZattooChannelGroup &group = m_channelGroups[channelItem["group_index"].GetInt()];
+    group.channels.insert(group.channels.end(), channel);
+    allChannels[cid] = channel;
+    m_channelsByCid[channel.cid] = channel;
+    m_channelsByUid[channel.iUniqueId] = channel;
   }
 
   PVRZattooChannelGroup favGroup;
@@ -428,10 +405,12 @@ PVR_ERROR ZatData::GetChannels(bool radio, kodi::addon::PVRChannelsResultSet& re
   return PVR_ERROR_NO_ERROR;
 }
 
-std::string ZatData::GetStreamUrl(std::string& jsonString, std::vector<kodi::addon::PVRStreamProperty>& properties) {
-  Document doc;
-  doc.Parse(jsonString.c_str());
-  if (doc.GetParseError() || !doc.HasMember("stream"))
+bool ZatData::IsDrmLimitApplied(Document& doc) {
+  return doc.HasMember("drm_limit_applied") && doc["drm_limit_applied"].GetBool();
+}
+
+std::string ZatData::GetStreamUrl(Document& doc, std::vector<kodi::addon::PVRStreamProperty>& properties) {
+  if (!doc.HasMember("stream"))
   {
     return "";
   }
@@ -442,11 +421,9 @@ std::string ZatData::GetStreamUrl(std::string& jsonString, std::vector<kodi::add
     const Value& watchUrl = (*itr);
     kodi::Log(ADDON_LOG_DEBUG, "Selected url for maxrate: %d", watchUrl["maxrate"].GetInt());
     url = Utils::JsonStringOrEmpty(watchUrl, "url");
-    if (m_settings->GetStreamType() == DASH_WIDEVINE) {
-      std::string licenseUrl = Utils::JsonStringOrEmpty(watchUrl, "license_url");
-      properties.emplace_back("inputstream.adaptive.license_key", licenseUrl + "||A{SSM}|");
-      properties.emplace_back("inputstream.adaptive.license_type", "com.widevine.alpha");
-    }
+    std::string licenseUrl = Utils::JsonStringOrEmpty(watchUrl, "license_url");
+    properties.emplace_back("inputstream.adaptive.license_key", licenseUrl + "||A{SSM}|");
+    properties.emplace_back("inputstream.adaptive.license_type", "com.widevine.alpha");
     break;
   }
   kodi::Log(ADDON_LOG_DEBUG, "Got url: %s", url.c_str());
@@ -460,14 +437,35 @@ PVR_ERROR ZatData::GetChannelStreamProperties(const kodi::addon::PVRChannel& cha
 
   ZatChannel* ownChannel = FindChannel(channel.GetUniqueId());
   kodi::Log(ADDON_LOG_DEBUG, "Get live url for channel %s", ownChannel->cid.c_str());
+  
+  bool forceWithoutDrm = GetDrmLevel() <= 0;
+  Document doc;
+  
+  while (true) {
+    std::ostringstream dataStream;
+    bool requiresDrm;
+    dataStream << GetQualityStreamParameter(ownChannel->cid, forceWithoutDrm, requiresDrm);
+    dataStream << GetBasicStreamParameters(requiresDrm) << "&format=json&timeshift=10800";
+    kodi::Log(ADDON_LOG_INFO, "Stream properties: %s.", dataStream.str().c_str());
+    int statusCode;
+    std::string jsonString = m_httpClient->HttpPost(m_session->GetProviderUrl() + "/zapi/watch/live/" + ownChannel->cid, dataStream.str(), statusCode);
+   
+    doc.Parse(jsonString.c_str());
+    if (doc.GetParseError())
+    {
+      return ret;
+    }
 
-  std::ostringstream dataStream;
-  dataStream << GetStreamParameters() << "&format=json&timeshift=10800";
+    if (forceWithoutDrm || !IsDrmLimitApplied(doc)) {
+      break;
+    }
+    forceWithoutDrm = true;
+    kodi::Log(ADDON_LOG_INFO, "Fallback to no-drm version.");
+    doc.SetNull();
+    doc.GetAllocator().Clear();
+  }
 
-  int statusCode;
-  std::string jsonString = m_httpClient->HttpPost(m_session->GetProviderUrl() + "/zapi/watch/live/" + ownChannel->cid, dataStream.str(), statusCode);
-
-  std::string strUrl = GetStreamUrl(jsonString, properties);
+  std::string strUrl = GetStreamUrl(doc, properties);
   if (!strUrl.empty())
   {
     SetStreamProperties(properties, strUrl);
@@ -588,6 +586,9 @@ bool ZatData::ParseRecordingsTimers(const Value& recordings, std::map<int, ZatRe
           details.genre = "";
         }
         details.description = Utils::JsonStringOrEmpty(program, "d");
+        details.seriesNumber = program.HasMember("s_no") && !program["s_no"].IsNull() ? program["s_no"].GetInt() : EPG_TAG_INVALID_SERIES_EPISODE;
+        details.episodeNumber = program.HasMember("e_no") && !program["e_no"].IsNull() ? program["e_no"].GetInt() : EPG_TAG_INVALID_SERIES_EPISODE;
+
         detailsById.insert(
             std::pair<int, ZatRecordingDetails>(program["id"].GetInt(), details));
       }
@@ -636,16 +637,6 @@ PVR_ERROR ZatData::GetTimers(kodi::addon::PVRTimersResultSet& results)
     const Value& recording = (*itr);
     int programId = recording["program_id"].GetInt();
 
-    std::string cid = Utils::JsonStringOrEmpty(recording, "cid");
-    auto iterator = m_channelsByCid.find(cid);
-    if (iterator == m_channelsByCid.end())
-    {
-      kodi::Log(ADDON_LOG_ERROR, "Channel %s not found for recording: %i",
-          cid.c_str(), programId);
-      continue;
-    }
-    ZatChannel channel = iterator->second;
-
     auto detailIterator = detailsById.find(programId);
     bool hasDetails = detailIterator != detailsById.end();
 
@@ -672,7 +663,14 @@ PVR_ERROR ZatData::GetTimers(kodi::addon::PVRTimersResultSet& results)
       tag.SetState(PVR_TIMER_STATE_SCHEDULED);
       tag.SetTimerType(1);
       tag.SetEPGUid(static_cast<unsigned int>(recording["program_id"].GetInt()));
-      tag.SetClientChannelUid(channel.iUniqueId);
+      std::string cid = Utils::JsonStringOrEmpty(recording, "cid");
+      auto iterator = m_channelsByCid.find(cid);
+      if (iterator != m_channelsByCid.end())
+      {
+        ZatChannel channel = iterator->second;
+        tag.SetClientChannelUid(channel.iUniqueId);
+      }
+      
       if (genre)
       {
         tag.SetGenreSubType(genre & 0x0F);
@@ -691,25 +689,21 @@ PVR_ERROR ZatData::GetTimers(kodi::addon::PVRTimersResultSet& results)
     {
       const Value& recording = (*itr);
       int tvSeriesId = recording["tv_series_id"].GetInt();
-  
-      std::string cid = Utils::JsonStringOrEmpty(recording, "cid");
-      auto iterator = m_channelsByCid.find(cid);
-      if (iterator == m_channelsByCid.end())
-      {
-        kodi::Log(ADDON_LOG_ERROR, "Channel %s not found for series recording: %i",
-            cid.c_str(), tvSeriesId);
-        continue;
-      }
-      ZatChannel channel = iterator->second;
-    
-      //genre
+      
       kodi::addon::PVRTimer tag;
 
+      std::string cid = Utils::JsonStringOrEmpty(recording, "cid");
+      auto iterator = m_channelsByCid.find(cid);
+      if (iterator != m_channelsByCid.end())
+      {
+        ZatChannel channel = iterator->second;
+        tag.SetClientChannelUid(channel.iUniqueId);
+      }
+      
       tag.SetClientIndex(static_cast<unsigned int>(tvSeriesId));
       tag.SetTitle(Utils::JsonStringOrEmpty(recording, "title"));
       tag.SetState(PVR_TIMER_STATE_SCHEDULED);
       tag.SetTimerType(2);
-      tag.SetClientChannelUid(channel.iUniqueId);
       results.Add(tag);
     }
   }
@@ -796,7 +790,7 @@ PVR_ERROR ZatData::DeleteTimer(const kodi::addon::PVRTimer& timer, bool forceDel
     {
       const Value& recording = (*itr);
       
-      int seriesId = recording["tv_series_id"].GetInt();
+      unsigned int seriesId = recording["tv_series_id"].GetInt();
       
       if (seriesId == timer.GetClientIndex()) {
         recordingId = recording["id"].GetInt();
@@ -864,16 +858,6 @@ PVR_ERROR ZatData::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultS
     const Value& recording = (*itr);
     int programId = recording["program_id"].GetInt();
 
-    std::string cid = Utils::JsonStringOrEmpty(recording, "cid");
-    auto iterator = m_channelsByCid.find(cid);
-    if (iterator == m_channelsByCid.end())
-    {
-      kodi::Log(ADDON_LOG_ERROR, "Channel %s not found for recording: %i",
-          cid.c_str(), programId);
-      continue;
-    }
-    ZatChannel channel = iterator->second;
-
     auto detailIterator = detailsById.find(programId);
     bool hasDetails = detailIterator != detailsById.end();
 
@@ -895,13 +879,27 @@ PVR_ERROR ZatData::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultS
       tag.SetRecordingId(std::to_string(recording["id"].GetInt()));
       tag.SetTitle(Utils::JsonStringOrEmpty(recording, "title"));
       tag.SetEpisodeName(Utils::JsonStringOrEmpty(recording, "episode_title"));
-      tag.SetPlot(hasDetails ? detailIterator->second.description : "");
+      if (hasDetails) {
+        tag.SetPlot(detailIterator->second.description);
+        tag.SetSeriesNumber(detailIterator->second.seriesNumber);
+        tag.SetEpisodeNumber(detailIterator->second.episodeNumber);
+      }
 
       std::string imageToken = Utils::JsonStringOrEmpty(recording, "image_token");
       std::string imageUrl = Utils::GetImageUrl(imageToken);;
       tag.SetIconPath(imageUrl);
-      tag.SetChannelUid(channel.iUniqueId);
-      tag.SetChannelName(channel.name);
+
+      std::string cid = Utils::JsonStringOrEmpty(recording, "cid");
+      auto iterator = m_channelsByCid.find(cid);
+      if (iterator != m_channelsByCid.end())
+      {
+        ZatChannel channel = iterator->second;
+        tag.SetChannelUid(channel.iUniqueId);
+        tag.SetChannelName(channel.name);
+      } else {
+        tag.SetChannelName(cid);
+      }      
+      
       time_t endTime = Utils::StringToTime(
           Utils::JsonStringOrEmpty(recording, "end").c_str());
       tag.SetRecordingTime(startTime);
@@ -969,41 +967,93 @@ PVR_ERROR ZatData::GetRecordingsAmount(bool deleted, int& amount)
   return PVR_ERROR_NO_ERROR;
 }
 
-std::string ZatData::GetStreamParameters() {
+std::string ZatData::GetBasicStreamParameters(bool requiresDrm) {
   std::string params = m_settings->GetZatEnableDolby() ? "&enable_eac3=true" : "";
-  params += "&stream_type=" + GetStreamTypeString();
+  
+  params += "&stream_type=" + GetStreamTypeString(requiresDrm);
+  int drmLevel = GetDrmLevel();
+  if (drmLevel > 0) {
+    params += "&max_drm_lvl=" + std::to_string(drmLevel);
+  }
 
   if (!m_settings->GetParentalPin().empty()) {
     params += "&youth_protection_pin=" + m_settings->GetParentalPin();
   }
-
+   
   return params;
 }
 
-std::string ZatData::GetStreamTypeString() {
-  switch (m_settings->GetStreamType()) {
-    case HLS:
-      return "hls7";
-    case DASH_WIDEVINE:
-      return "dash_widevine";
-    default:
-      return "dash";
+std::string ZatData::GetQualityStreamParameter(const std::string& cid, bool forceWithoutDrm, bool& requiresDrm) {
+  requiresDrm = !forceWithoutDrm;
+  auto iterator = m_channelsByCid.find(cid);
+  if (iterator != m_channelsByCid.end())
+  {
+    ZatChannel channel = iterator->second;
+    std::string selectedQuality;
+    for (auto const& pair : channel.qualityWithDrm)
+    {
+      if (!forceWithoutDrm || !pair.second) {
+        selectedQuality = pair.first;
+        requiresDrm = pair.second;
+        break;
+      }
+    }
+
+    if (!selectedQuality.empty()) {
+      kodi::Log(ADDON_LOG_INFO, "Selected quality: %s, requiring drm: %s", selectedQuality.c_str(), requiresDrm ? "true" : "false");
+      return "&quality=" + selectedQuality;
+    }
   }
+  
+  return "";
+}
+
+int ZatData::GetDrmLevel() {
+  int drmLevel = m_settings->DrmLevel();
+  if (drmLevel == 0) {
+    return Utils::RunsOnLinux() ? 3 : 1;
+  } else {
+    return drmLevel;
+  }
+}
+
+std::string ZatData::GetStreamTypeString(bool withDrm) {
+    if (!withDrm) {
+      return "dash";
+    }
+    return "dash_widevine";
 }
 
 PVR_ERROR ZatData::GetRecordingStreamProperties(const kodi::addon::PVRRecording& recording,
                                                 std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
   kodi::Log(ADDON_LOG_DEBUG, "Get url for recording %s", recording.GetRecordingId().c_str());
-
-  std::ostringstream dataStream;
-  dataStream << GetStreamParameters();
-
-  int statusCode;
-  std::string jsonString = m_httpClient->HttpPost(m_session->GetProviderUrl() + "/zapi/watch/recording/" + recording.GetRecordingId(), dataStream.str(), statusCode);
-
-  std::string strUrl = GetStreamUrl(jsonString, properties);
   PVR_ERROR ret = PVR_ERROR_FAILED;
+  
+  std::string cid = "";
+  if (m_channelsByUid.count(recording.GetChannelUid())) {
+    ZatChannel& channel = m_channelsByUid[recording.GetChannelUid()];
+    cid = channel.cid;
+  }
+  
+  Document doc;
+  
+  bool useWidevine = GetDrmLevel() > -1;
+  
+  std::ostringstream dataStream;
+  dataStream << GetBasicStreamParameters(useWidevine);
+  kodi::Log(ADDON_LOG_INFO, "Stream properties: %s.", dataStream.str().c_str());
+  int statusCode;
+  
+  std::string jsonString = m_httpClient->HttpPost(m_session->GetProviderUrl() + "/zapi/watch/recording/" + recording.GetRecordingId(), dataStream.str(), statusCode);
+  
+  doc.Parse(jsonString.c_str());
+  if (doc.GetParseError())
+  {
+    return ret;
+  }
+                  
+  std::string strUrl = GetStreamUrl(doc, properties);
   if (!strUrl.empty())
   {
     SetStreamProperties(properties, strUrl);
@@ -1124,18 +1174,36 @@ PVR_ERROR ZatData::GetEPGTagStreamProperties(const kodi::addon::PVREPGTag& tag,
 
 std::string ZatData::GetStreamUrlForProgram(const std::string& cid, int programId, std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
-  std::ostringstream dataStream;
-
-  std::string jsonString;
-
   kodi::Log(ADDON_LOG_DEBUG, "Get timeshift url for channel %s and program %i", cid.c_str(), programId);
+  
+  bool forceWithoutDrm = GetDrmLevel() <= 0;
+  Document doc;
+  
+  while (true) {
+    std::ostringstream dataStream;
+    bool requiresDrm;
+    dataStream << GetQualityStreamParameter(cid, forceWithoutDrm, requiresDrm);
+    dataStream << GetBasicStreamParameters(requiresDrm);
+    dataStream << "&pre_padding=0&post_padding=0";
+    kodi::Log(ADDON_LOG_INFO, "Stream properties: %s.", dataStream.str().c_str());
+    int statusCode;
+    std::string jsonString = m_httpClient->HttpPost(m_session->GetProviderUrl() + "/zapi/v3/watch/replay/" + cid + "/" + std::to_string(programId), dataStream.str(), statusCode);
+    doc.Parse(jsonString.c_str());
+    if (doc.GetParseError())
+    {
+      return "";
+    }
 
-  dataStream << GetStreamParameters();
-  dataStream << "&pre_padding=0&post_padding=0";
-  int statusCode;
-  jsonString = m_httpClient->HttpPost(m_session->GetProviderUrl() + "/zapi/v3/watch/replay/" + cid + "/" + std::to_string(programId), dataStream.str(), statusCode);
-
-  std::string strUrl = GetStreamUrl(jsonString, properties);
+    if (forceWithoutDrm || !IsDrmLimitApplied(doc)) {
+      break;
+    }
+    forceWithoutDrm = true;
+    kodi::Log(ADDON_LOG_INFO, "Fallback to no-drm version.");
+    doc.SetNull();
+    doc.GetAllocator().Clear();
+  }
+  
+  std::string strUrl = GetStreamUrl(doc, properties);
   return strUrl;
 }
 
@@ -1172,7 +1240,7 @@ bool ZatData::SessionInitialized()
   if (m_epgProvider) {
     delete m_epgProvider;
   }
-  kodi::Log(ADDON_LOG_INFO, "Stream type: %s", GetStreamTypeString().c_str());
+  kodi::Log(ADDON_LOG_INFO, "DRM Level: %i", m_settings->DrmLevel());
   if (!LoadChannels()) {
     return false;
   }
